import doctest
import unittest

from pkg_resources import resource_listdir

import zope.testbrowser.wsgi
from zope.app.wsgi.testlayer import http

import grokcore.traverser


class Layer(
        zope.testbrowser.wsgi.TestBrowserLayer,
        zope.app.wsgi.testlayer.BrowserLayer):
    pass


layer = Layer(grokcore.traverser)


def http_call(method, path, data=None, **kw):
    """Function to help make RESTful calls.

    method - HTTP method to use
    path - testbrowser style path
    data - (body) data to submit
    kw - any request parameters
    """

    if path.startswith('http://localhost'):
        path = path[len('http://localhost'):]
    request_string = '{} {} HTTP/1.1\n'.format(method, path)
    for key, value in kw.items():
        request_string += '{}: {}\n'.format(key, value)
    if data is not None:
        request_string += '\r\n'
        request_string += data
    return http(request_string, handle_errors=False)


def suiteFromPackage(name):
    layer_dir = 'functional'
    files = resource_listdir(__name__, f'{layer_dir}/{name}')
    suite = unittest.TestSuite()
    for filename in files:
        if not filename.endswith('.py'):
            continue
        if filename == '__init__.py':
            continue

        dottedname = 'grokcore.traverser.tests.{}.{}.{}'.format(
            layer_dir, name, filename[:-3])
        test = doctest.DocTestSuite(
            dottedname,
            extraglobs=dict(http_call=http_call,
                            http=http,
                            getRootFolder=layer.getRootFolder),
            optionflags=(doctest.ELLIPSIS +
                         doctest.NORMALIZE_WHITESPACE +
                         doctest.REPORT_NDIFF +
                         doctest.IGNORE_EXCEPTION_DETAIL))
        test.layer = layer

        suite.addTest(test)
    return suite


def test_suite():
    suite = unittest.TestSuite()
    for name in ['traversal']:
        suite.addTest(suiteFromPackage(name))
    return suite



"""
Containers can have an explicit traverser associated with them.  The
behaviour falls back to basic container traversal if the 'traverse'
method returns None. Normal behaviour also means that the standard
Zope 3 paradigm "items before views" is supported in the fallback.

  >>> getRootFolder()["herd"] = herd = Herd()
  >>> herd['manfred'] = Mammoth('Manfred')
  >>> herd['ellie'] = Mammoth('Ellie')

  >>> from zope.testbrowser.wsgi import Browser
  >>> browser = Browser()
  >>> browser.handleErrors = False

When we look up 'manfred', we'll get the Mammoth object as expected:

  >>> browser.open("http://localhost/herd/manfred")
  >>> print(browser.contents)
  Hello Manfred

When we look up 'ellie', we also get a Mammoth object and not the
Ellie view:

  >>> browser.open("http://localhost/herd/ellie")
  >>> print(browser.contents)
  Hello Ellie

We can, of course, get to the Ellie view explicitly:

  >>> browser.open("http://localhost/herd/@@ellie")
  >>> print(browser.contents)
  Hi, it's me, the Ellie view!

"""
import grokcore.component as grok
import grokcore.content as content
import grokcore.view as view

import grokcore.traverser


class Herd(content.Container):
    pass


class Traverser(grokcore.traverser.Traverser):
    grok.context(Herd)

    def traverse(self, name):
        # we don't really need to do anything here as we want to test
        # the fallback behaviour
        pass


class Ellie(view.View):
    grok.context(Herd)
    grok.name('ellie')

    def render(self):
        return "Hi, it's me, the Ellie view!"


class Mammoth(content.Model):
    def __init__(self, name):
        self.name = name


class MammothIndex(view.View):
    grok.context(Mammoth)
    grok.name('index')

    def render(self):
        return "Hello " + self.context.name.title()



##############################################################################
#
# Copyright (c) 2006-2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Grok interfaces
"""
import grokcore.component.interfaces
import grokcore.security.interfaces
import grokcore.view.interfaces
from zope import interface
from zope.publisher.interfaces.http import IHTTPRequest


class IBaseClasses(
        grokcore.component.interfaces.IBaseClasses,
        grokcore.security.interfaces.IBaseClasses,
        grokcore.view.interfaces.IBaseClasses):

    Traverser = interface.Attribute("Base class for custom traversers.")


class IGrokTraverser(
        grokcore.component.interfaces.IGrokcoreComponentAPI,
        grokcore.security.interfaces.IGrokcoreSecurityAPI,
        grokcore.view.interfaces.IGrokcoreViewAPI,
        IBaseClasses):
    pass


class IRESTLayer(IHTTPRequest):
    """REST-specific Request functionality.

    Base Interfaces for defining REST-layers.
    """



import os

from setuptools import find_packages
from setuptools import setup


def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()


long_description = (
    read('README.rst')
    + '\n' +
    read('CHANGES.rst')
)


tests_require = [
    'grokcore.content',
    'grokcore.view [test]',
    'zope.app.appsetup',
    'zope.app.wsgi',
    'zope.login',
    'zope.principalregistry',
    'zope.securitypolicy',
    'zope.testbrowser',
    'zope.testing',
]


setup(
    name='grokcore.traverser',
    version='4.0',
    author='Grok Team',
    author_email='zope-dev@zope.dev',
    url='https://github.com/zopefoundation/grokcore.traverser',
    description='Traverser for the Grok Framework',
    long_description=long_description,
    license='ZPL',
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Zope Public License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: Implementation',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Framework :: Zope :: 3',
    ],
    packages=find_packages('src'),
    package_dir={'': 'src'},
    namespace_packages=['grokcore'],
    include_package_data=True,
    zip_safe=False,
    python_requires='>=3.7',
    install_requires=[
        'grokcore.component >= 2.5',
        'grokcore.security',
        'grokcore.view',
        'martian',
        'setuptools',
        'zope.component',
        'zope.interface',
        'zope.publisher',
    ],
    tests_require=tests_require,
    extras_require={'test': tests_require},
)



##############################################################################
#
# Copyright (c) 2006-2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import martian


class traversable(martian.Directive):
    """The `grok.traversable()` directive.

    Each time this directive is used inside of a class, it designates an
    attribute of that class which URLs should be able to traverse.  For
    example, the declaration:

        class Mammoth(grok.Model):
            grok.traversable('thighbone')

    means that if the URL `/app/mymammoth` designates a Mammoth, then
    `/app/mymammoth/thighbone` will also be a valid URL (assuming that
    the Mammoth instance, at runtime, indeed has an attribute by that
    name)!  By default, the name that must be appended to the URL should
    simply be the same as the name of the attribute; but by providing a
    `name` keyword argument, the programmer can designate another name
    to appear in the URL instead of the raw attribute name.

    """
    scope = martian.CLASS
    store = martian.DICT

    def factory(self, attr, name=None):
        if name is None:
            name = attr
        return (name, attr)
